/*++
Copyright (c) Realtek Semiconductor Corp. All rights reserved.

Module Name:
	Hal88XXGen.c
	
Abstract:
	Defined RTL8192EE HAL Function
	    
Major Change History:
	When       Who               What
	---------- ---------------   -------------------------------
	2012-03-23 Filen            Create.	
--*/

#ifndef __ECOS
#include "HalPrecomp.h"
#else
#include "../../../HalPrecomp.h"
#endif
#include "../../../../8192cd.h"
typedef enum _RTL8192E_C2H_EVT
{
	C2H_8192E_DBG = 0,
	C2H_8192E_LB = 1,
	C2H_8192E_TXBF = 2,
	C2H_8192E_TX_REPORT = 3,
	C2H_8192E_TX_RATE =4,
	C2H_8192E_BT_INFO = 9,
	C2H_8192E_BT_MP = 11,
	MAX_8192E_C2HEVENT
}RTL8192E_C2H_EVT;

RT_STATUS
InitPON8192EE(
    IN  HAL_PADAPTER Adapter,
    IN  u4Byte     	ClkSel        
)
{
    u32     bytetmp;
    u32     retry;
    
    RT_TRACE_F( COMP_INIT, DBG_LOUD, ("\n"));

	HAL_RTL_W8(REG_RSV_CTRL, 0x00);

    // Add by Eric 2013/01/24
    // For 92E MP chip, power on sometimes crystal clk setting error
    // clk set 25M, value 0x00
	if(ClkSel == XTAL_CLK_SEL_25M) {
		HAL_RTL_W16(REG_AFE_CTRL4, 0x002a);
		HAL_RTL_W8(REG_AFE_CTRL2, 5);
	} else if (ClkSel == XTAL_CLK_SEL_40M){
		HAL_RTL_W16(REG_AFE_CTRL4, 0x002a);
		HAL_RTL_W8(REG_AFE_CTRL2, 1);
	}
	
	//pattern Patching EQC fail IC(AFE issue)
	//Setting is suggested by SD1-Pisa
	//Set 0x78[21] = 0, 0x28[6] = 0
	//0x28[6] = 0 is done by XTAL seletion 0x28[6] = 0
	HAL_RTL_W32(REG_AFE_CTRL4, HAL_RTL_R32(REG_AFE_CTRL4) & ~BIT21);
	
	if (!HalPwrSeqCmdParsing88XX(Adapter, PWR_CUT_ALL_MSK, PWR_FAB_ALL_MSK,
			PWR_INTF_PCI_MSK, rtl8192E_card_enable_flow))
    {
        RT_TRACE( COMP_INIT, DBG_SERIOUS, ("%s %d, HalPwrSeqCmdParsing init fail!!!\n", __FUNCTION__, __LINE__));
        return RT_STATUS_FAILURE;
    }

	// Power on when re-enter from IPS/Radio off/card disable
    HAL_RTL_W8(REG_AFE_CTRL1, HAL_RTL_R8(REG_AFE_CTRL1) | BIT0);

    bytetmp = HAL_RTL_R16(REG_SYS_PW_CTRL);
    bytetmp &= 0xE7ff;
    bytetmp |= 0x0800;
    HAL_RTL_W16(REG_SYS_PW_CTRL, bytetmp);

    while (!((bytetmp = HAL_RTL_R32(REG_SYS_PW_CTRL)) & 0x00020000));

    bytetmp = HAL_RTL_R16(REG_SYS_PW_CTRL);
    bytetmp &= 0x7FFF;
    HAL_RTL_W16(REG_SYS_PW_CTRL, bytetmp);

    bytetmp = HAL_RTL_R16(REG_SYS_PW_CTRL);
    bytetmp &= 0xE7ff;
    bytetmp |= 0x0000;
    HAL_RTL_W16(REG_SYS_PW_CTRL, bytetmp);

    HAL_delay_ms(1);

	// auto enable WLAN
	// Power On Reset for MAC Block
	bytetmp = HAL_RTL_R8(REG_SYS_PW_CTRL+1) | BIT(0);
	HAL_delay_us(2);
	HAL_RTL_W8(REG_SYS_PW_CTRL+1, bytetmp);
	HAL_delay_us(2);

	bytetmp = HAL_RTL_R8(REG_SYS_PW_CTRL+1);
	HAL_delay_us(2);
	retry = 0;
	while((bytetmp & BIT(0)) && retry < 1000){
		retry++;
		HAL_delay_us(50);
		bytetmp = HAL_RTL_R8(REG_SYS_PW_CTRL+1);
		HAL_delay_us(50);
	}
	
    RT_TRACE(COMP_INIT, DBG_WARNING, ("%s: RTL_R8(APS_FSMCO+1) retry times=%d\n", (char *)__FUNCTION__, retry) );

	{
		int val;
		val = HAL_RTL_R32(REG_SYS_PW_CTRL);
        RT_TRACE(COMP_INIT, DBG_TRACE, ("FSMCO11=0x%x\n", val) );
	}
	
	if (bytetmp & BIT(0)) {
        RT_TRACE(COMP_INIT, DBG_SERIOUS, ("%s ERROR: auto enable WLAN failed!!(0x%02X)\n", __FUNCTION__, bytetmp) );
	}	

    HAL_RTL_W16(REG_SYS_FUNC_EN, HAL_RTL_R16(REG_SYS_FUNC_EN) & ~BIT_FEN_CPUEN);
    
    HAL_delay_us(2);

    // check LDO mode 
    if(HAL_RTL_R32(REG_SYS_CFG1)&BIT24) {
        // LDO mode set 0x7C
        HAL_RTL_W8(REG_LDO_SWR_CTRL,0xc3);
    } else {
        // SPS mode
        HAL_RTL_W8(REG_LDO_SWR_CTRL,0x83);
    }

    return  RT_STATUS_SUCCESS;    
}


RT_STATUS
StopHW8192EE(
    IN  HAL_PADAPTER Adapter
)
{
    // TODO:

    return RT_STATUS_SUCCESS;
}


RT_STATUS	
hal_Associate_8192EE(
    HAL_PADAPTER        Adapter,
    BOOLEAN                 IsDefaultAdapter
)
{
    PHAL_INTERFACE              pHalFunc = GET_HAL_INTERFACE(Adapter);
    PHAL_DATA_TYPE              pHalData = _GET_HAL_DATA(Adapter);

    //
    //Initialization Related
    //
    pHalData->AccessSwapCtrl        = HAL_ACCESS_SWAP_MEM;
    pHalFunc->InitPONHandler        = InitPON8192EE;
    pHalFunc->InitMACHandler        = InitMAC88XX;
    pHalFunc->InitFirmwareHandler   = InitFirmware88XX;
    pHalFunc->InitHCIDMAMemHandler  = InitHCIDMAMem88XX;
    pHalFunc->InitHCIDMARegHandler  = InitHCIDMAReg88XX;    
#if CFG_HAL_SUPPORT_MBSSID    
    pHalFunc->InitMBSSIDHandler     = InitMBSSID88XX;
#endif  //CFG_HAL_SUPPORT_MBSSID
    pHalFunc->InitVAPIMRHandler     = InitVAPIMR88XX;


    //
    //Stop Related
    //
#if CFG_HAL_SUPPORT_MBSSID        
    pHalFunc->StopMBSSIDHandler     = StopMBSSID88XX;
#endif  //CFG_HAL_SUPPORT_MBSSID
    pHalFunc->StopHWHandler         = StopHW88XX;
    pHalFunc->StopSWHandler         = StopSW88XX;
    pHalFunc->DisableVXDAPHandler   = DisableVXDAP88XX;


    //
    //ISR Related
    //
    pHalFunc->InitIMRHandler                    = InitIMR88XX;
    pHalFunc->EnableIMRHandler                  = EnableIMR88XX;
    pHalFunc->InterruptRecognizedHandler        = InterruptRecognized88XX;
    pHalFunc->GetInterruptHandler               = GetInterrupt88XX;
    pHalFunc->AddInterruptMaskHandler           = AddInterruptMask88XX;
    pHalFunc->RemoveInterruptMaskHandler        = RemoveInterruptMask88XX;
    pHalFunc->DisableRxRelatedInterruptHandler  = DisableRxRelatedInterrupt88XX;
    pHalFunc->EnableRxRelatedInterruptHandler   = EnableRxRelatedInterrupt88XX;


    //
    //Tx Related
    //
    pHalFunc->PrepareTXBDHandler            = PrepareTXBD88XX;    
    pHalFunc->FillTxHwCtrlHandler           = FillTxHwCtrl88XX;
    pHalFunc->SyncSWTXBDHostIdxToHWHandler  = SyncSWTXBDHostIdxToHW88XX;
    pHalFunc->TxPollingHandler              = TxPolling88XX;
    pHalFunc->SigninBeaconTXBDHandler       = SigninBeaconTXBD88XX;
    pHalFunc->SetBeaconDownloadHandler      = SetBeaconDownload88XX;
    pHalFunc->FillBeaconDescHandler         = FillBeaconDesc88XX;
    pHalFunc->GetTxQueueHWIdxHandler        = GetTxQueueHWIdx88XX;
    pHalFunc->MappingTxQueueHandler         = MappingTxQueue88XX;
    pHalFunc->QueryTxConditionMatchHandler  = QueryTxConditionMatch88XX;
#if CFG_HAL_TX_SHORTCUT
//    pHalFunc->GetShortCutTxDescHandler      = GetShortCutTxDesc88XX;
//    pHalFunc->ReleaseShortCutTxDescHandler  = ReleaseShortCutTxDesc88XX;
    pHalFunc->GetShortCutTxBuffSizeHandler  = GetShortCutTxBuffSize88XX;
    pHalFunc->SetShortCutTxBuffSizeHandler  = SetShortCutTxBuffSize88XX;
    pHalFunc->CopyShortCutTxDescHandler     = CopyShortCutTxDesc88XX;
    pHalFunc->FillShortCutTxHwCtrlHandler   = FillShortCutTxHwCtrl88XX;    
#endif // CFG_HAL_TX_SHORTCUT

    //
    //Rx Related
    //
    pHalFunc->PrepareRXBDHandler            = PrepareRXBD88XX;
    pHalFunc->QueryRxDescHandler            = QueryRxDesc88XX;
    pHalFunc->UpdateRXBDInfoHandler         = UpdateRXBDInfo88XX;
    pHalFunc->UpdateRXBDHWIdxHandler        = UpdateRXBDHWIdx88XX;
    pHalFunc->UpdateRXBDHostIdxHandler      = UpdateRXBDHostIdx88XX;    

    //
    // General operation
    //
    pHalFunc->GetChipIDMIMOHandler          =   GetChipIDMIMO88XX;
    pHalFunc->SetHwRegHandler               =   SetHwReg88XX;
    pHalFunc->GetHwRegHandler               =   GetHwReg88XX;
    pHalFunc->SetMACIDSleepHandler          =   SetMACIDSleep88XX;
	pHalFunc->CheckHangHandler              =   CheckHang88XX;

    //
    // Security Related     
    //
    pHalFunc->CAMReadMACConfigHandler       =   CAMReadMACConfig88XX;
    pHalFunc->CAMEmptyEntryHandler          =   CAMEmptyEntry88XX;
    pHalFunc->CAMFindUsableHandler          =   CAMFindUsable88XX;
    pHalFunc->CAMProgramEntryHandler        =   CAMProgramEntry88XX;


    //
    // PHY/RF Related
    //
    pHalFunc->PHYSetCCKTxPowerHandler       = PHYSetCCKTxPower88XX_N;
    pHalFunc->PHYSetOFDMTxPowerHandler      = PHYSetOFDMTxPower88XX_N;
    pHalFunc->PHYSwBWModeHandler            = SwBWMode88XX_N;
    //pHalFunc->TXPowerTrackingHandler        = TXPowerTracking_ThermalMeter_88XX;    


    //
    // Firmware CMD IO related
    //
    pHalData->H2CBufPtr88XX     = 0;
    pHalData->bFWReady          = _FALSE;
    pHalFunc->FillH2CCmdHandler             = FillH2CCmd88XX;
    pHalFunc->UpdateHalRAMaskHandler        = UpdateHalRAMask88XX;
    pHalFunc->UpdateHalMSRRPTHandler        = UpdateHalMSRRPT88XX;
    pHalFunc->SetAPOffloadHandler           = SetAPOffload88XX;
    pHalFunc->SetRsvdPageHandler	        = SetRsvdPage88XX;
    pHalFunc->GetRsvdPageLocHandler	        = GetRsvdPageLoc88XX;
    pHalFunc->DownloadRsvdPageHandler	    = DownloadRsvdPage88XX;
    pHalFunc->C2HHandler                    = C2HHandler88XX;
    pHalFunc->DumpRxBDescTestHandler = DumpRxBDesc88XX;
    pHalFunc->DumpTxBDescTestHandler = DumpTxBDesc88XX;
    
    return  RT_STATUS_SUCCESS;    
}


void 
InitMAC8192EE(
    IN  HAL_PADAPTER Adapter
)
{














    
}
#ifdef BEAMFORMING_SUPPORT

VOID
SetBeamformRfMode92E(
	struct rtl8192cd_priv *priv,
	PRT_BEAMFORMING_INFO 	pBeamformingInfo
	)
	{
		u1Byte					i;
		BOOLEAN 				bSelfBeamformer = FALSE;
		BOOLEAN 				bSelfBeamformee = FALSE;
		RT_BEAMFORMING_ENTRY	BeamformEntry;
		BEAMFORMING_CAP 	BeamformCap = BEAMFORMING_CAP_NONE;
	
	
		for(i = 0; i < BEAMFORMING_ENTRY_NUM; i++)
		{
			BeamformEntry = pBeamformingInfo->BeamformingEntry[i];
		
			if(BeamformEntry.bUsed)
			{
				if( (BeamformEntry.BeamformEntryCap & BEAMFORMEE_CAP_VHT_SU) ||
					(BeamformEntry.BeamformEntryCap & BEAMFORMEE_CAP_HT_EXPLICIT))
					bSelfBeamformee = TRUE;
				if( (BeamformEntry.BeamformEntryCap & BEAMFORMER_CAP_VHT_SU) ||
					(BeamformEntry.BeamformEntryCap & BEAMFORMER_CAP_HT_EXPLICIT))
					bSelfBeamformer = TRUE;
			}
	
			if(bSelfBeamformer && bSelfBeamformee)
				i = BEAMFORMING_ENTRY_NUM;
		}
	
		if(bSelfBeamformer)
			BeamformCap |= BEAMFORMER_CAP;
		if(bSelfBeamformee)
			BeamformCap |= BEAMFORMEE_CAP;
	
		if(BeamformCap == pBeamformingInfo->BeamformCap)
			return;
		else 
			pBeamformingInfo->BeamformCap = BeamformCap;
	
		if(get_rf_mimo_mode(priv) == MIMO_1T1R)
			return;
	
		PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_WE_LUT, 0x80000,0x1); // RF Mode table write enable
		PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_WE_LUT, 0x80000,0x1); // RF Mode table write enable
		if(bSelfBeamformer)
		{ 
			// Paath_A
			PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_RCK_OS, 0xfffff,0x18000); // Select RX mode	0x30=0x18000
			PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_TXPA_G1, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_TXPA_G2, 0xfffff,0x77fc2); // Enable TXIQGEN in RX mode
			// Path_B
			PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_RCK_OS, 0xfffff,0x18000); // Select RX mode
			PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_TXPA_G1, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_TXPA_G2, 0xfffff,0x77fc2); // Enable TXIQGEN in RX mode
		}
		else
		{
			// Paath_A
			PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_RCK_OS, 0xfffff,0x18000); // Select RX mode
			PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_TXPA_G1, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_TXPA_G2, 0xfffff,0x77f82); // Disable TXIQGEN in RX mode
			// Path_B
			PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_RCK_OS, 0xfffff,0x18000); // Select RX mode
			PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_TXPA_G1, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_TXPA_G2, 0xfffff,0x77f82); // Disable TXIQGEN in RX mode
		}
#if 0
		if(bSelfBeamformer)
		{
			// Paath_A
			PHY_SetRFReg(Adapter, ODM_RF_PATH_A, RF_ModeTableAddr, 0xfffff,0x10000); // Select TX mode	0x30=0x10000
			PHY_SetRFReg(Adapter, ODM_RF_PATH_A, RF_ModeTableData0, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(Adapter, ODM_RF_PATH_A, RF_ModeTableData1, 0xfffff,0x17e7e); // Enable RXIQGEN in TX mode
			// Path_B
			PHY_SetRFReg(Adapter, ODM_RF_PATH_B, RF_ModeTableAddr, 0xfffff,0x10000); // Select TX mode
			PHY_SetRFReg(Adapter, ODM_RF_PATH_B, RF_ModeTableData0, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(Adapter, ODM_RF_PATH_B, RF_ModeTableData1, 0xfffff,0x17e7e); // Enable RXIQGEN in TX mode 
		}
		else  
		{
			// Paath_A
			PHY_SetRFReg(Adapter, ODM_RF_PATH_A, RF_ModeTableAddr, 0xfffff,0x10000); // Select TX mode
			PHY_SetRFReg(Adapter, ODM_RF_PATH_A, RF_ModeTableData0, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(Adapter, ODM_RF_PATH_A, RF_ModeTableData1, 0xfffff,0x07e7e); // Disable RXIQGEN in TX mode
			// Path_B
			PHY_SetRFReg(Adapter, ODM_RF_PATH_B, RF_ModeTableAddr, 0xfffff,0x10000); // Select RX mode
			PHY_SetRFReg(Adapter, ODM_RF_PATH_B, RF_ModeTableData0, 0xfffff,0x0000f); // Set Table data
			PHY_SetRFReg(Adapter, ODM_RF_PATH_B, RF_ModeTableData1, 0xfffff,0x07e7e); // Disable TXIQGEN in TX mode
		}
#endif
		PHY_SetRFReg(priv, ODM_RF_PATH_A, RF_WE_LUT, 0x80000,0x0); // RF Mode table write disable
		PHY_SetRFReg(priv, ODM_RF_PATH_B, RF_WE_LUT, 0x80000,0x0); // RF Mode table write disable
	
		if(bSelfBeamformer){
			PHY_SetBBReg(priv, 0x90c, 0xffffffff, 0x83321333);
			PHY_SetBBReg(priv, 0xa04, BIT31|BIT30, 0x3);
		}
		else
			PHY_SetBBReg(priv, 0x90c, 0xffffffff, 0x81121313);
	}



VOID
SetBeamformEnter92E(
	struct rtl8192cd_priv *priv,
	u1Byte				Idx
	)
{
	u1Byte					i = 0;
	u4Byte					CSI_Param;	
	PRT_BEAMFORMING_INFO 	pBeamformingInfo = &(priv->pshare->BeamformingInfo);
	RT_BEAMFORMING_ENTRY	BeamformEntry = pBeamformingInfo->BeamformingEntry[Idx];
	
	u2Byte					STAid = 0;


	SetBeamformRfMode92E(priv, pBeamformingInfo);


	if(OPMODE & WIFI_ADHOC_STATE)
		STAid = BeamformEntry.AID;
	else 
		STAid = BeamformEntry.P_AID;
	
//	panic_printk("%s, STAid=%d\n", __FUNCTION__, STAid);

	if (IS_TEST_CHIP(priv))
		RTL_W8( REG_SND_PTCL_CTRL_8192E, 0x1B); 		// Disable SIG-B CRC8 check
	else
		RTL_W8( REG_SND_PTCL_CTRL_8192E, 0xCB);	
	
	// MAC addresss/Partial AID of Beamformer
	if(Idx == 0)
	{
		for(i = 0; i < 6 ; i++)
			RTL_W8( (REG_BFMER0_INFO_8192E+i), BeamformEntry.MacAddr[i]);
		
		RTL_W16( REG_BFMER0_INFO_8192E+6, BeamformEntry.P_AID);
	}
	else
	{
		for(i = 0; i < 6 ; i++)
			RTL_W8( (REG_BFMER1_INFO_8192E+i), BeamformEntry.MacAddr[i]);

		RTL_W16( REG_BFMER1_INFO_8192E+6, BeamformEntry.P_AID);
	}

	// CSI report parameters of Beamformer
	
		CSI_Param = 0x03090309;//Nc =2, V matrix

	if(get_rf_mimo_mode(priv) == MIMO_2T2R)
		RTL_W32(0xD80, 0x01081008);

	RTL_W32( REG_CSI_RPT_PARAM_BW20_92E, CSI_Param);
	RTL_W32( REG_CSI_RPT_PARAM_BW40_92E, CSI_Param);
	//RTL_W32( REG_CSI_RPT_PARAM_BW80_8812, CSI_Param);

	// P_AID of Beamformee & enable NDPA transmission
	if(Idx == 0)
	{	
		RTL_W16( REG_TXBF_CTRL, STAid);	
		RTL_W8( REG_TXBF_CTRL+3, RTL_R8( REG_TXBF_CTRL+3)|BIT6|BIT7|BIT4);
	}	
	else
	{
		RTL_W16( REG_TXBF_CTRL+2, STAid |BIT14| BIT15|BIT12);
	}	

	// CSI report parameters of Beamformee
	if(Idx == 0)	
	{
		// Get BIT24 & BIT25
		u1Byte	tmp = RTL_R8( REG_BFMEE_SEL_8192E+3) & 0x3;	
		RTL_W8( REG_BFMEE_SEL_8192E+3, tmp | 0x60);
		RTL_W16( REG_BFMEE_SEL_8192E, STAid | BIT9);
	}	
	else
	{
		// Set BIT25
		RTL_W16( REG_BFMEE_SEL_8192E+2, STAid | 0xE2);
	}

	// Timeout value for MAC to leave NDP_RX_standby_state 60 us
//	RTL_W8( REG_SND_PTCL_CTRL_8812+3, 0x3C);
	RTL_W8( REG_SND_PTCL_CTRL_8192E+3, 0x50);				// // ndp_rx_standby_timer


//	if(pHalData->bIsMPChip == FALSE) 
	if (IS_TEST_CHIP(priv))		
	{
		// VHT category value 
		RTL_W8( REG_SND_PTCL_CTRL_8192E+1, ACT_CAT_VHT);
		// NDPA subtype
		RTL_W8( REG_SND_PTCL_CTRL_8192E+2, Type_NDPA >> 4);
	}	
}


VOID
SetBeamformLeave92E(
	struct rtl8192cd_priv *priv,
	u1Byte				Idx
	)
{
//	RT_BEAMFORMING_INFO	BeamformInfo = Adapter->MgntInfo.BeamformingInfo[Idx];

	/*	Clear P_AID of Beamformee
	* 	Clear MAC addresss of Beamformer
	*	Clear Associated Bfmee Sel
	*/	
//	panic_printk("[%d][%s]\n",__LINE__,__FUNCTION__);
	if(Idx == 0)
	{	
		RTL_W16( REG_TXBF_CTRL, 0);	
		RTL_W32( REG_BFMER0_INFO_8192E, 0);
		RTL_W16( REG_BFMER0_INFO_8192E+4, 0);
		RTL_W16( REG_BFMEE_SEL_8192E, 0);
	}	
	else
	{
		RTL_W16( REG_TXBF_CTRL+2, RTL_R16( REG_TXBF_CTRL+2) & 0xF000);
		RTL_W32( REG_BFMER1_INFO_8192E, 0);
		RTL_W16( REG_BFMER1_INFO_8192E+4, 0);
		RTL_W16( REG_BFMEE_SEL_8192E+2, RTL_R16( REG_BFMEE_SEL_8192E+2) & 0x60);
	}	
}



VOID
SetBeamformStatus92E(
	struct rtl8192cd_priv *priv,
	 u1Byte				Idx
	)
{
	u2Byte					BeamCtrlVal;
	u4Byte					BeamCtrlReg;
//	PRT_BEAMFORMING_INFO 	pBeamformingInfo = GET_BEAMFORM_INFO(&(Adapter->MgntInfo));
	PRT_BEAMFORMING_INFO pBeamformingInfo = &(priv->pshare->BeamformingInfo);
	RT_BEAMFORMING_ENTRY	BeamformEntry = pBeamformingInfo->BeamformingEntry[Idx];

//	if(ACTING_AS_IBSS(Adapter))
	if(OPMODE & WIFI_ADHOC_STATE)
		BeamCtrlVal = BeamformEntry.MacId;
	else 
		BeamCtrlVal = BeamformEntry.P_AID;

	if(Idx == 0)
		BeamCtrlReg = REG_TXBF_CTRL;
	else
	{
		BeamCtrlReg = REG_TXBF_CTRL+2;
		BeamCtrlVal |= BIT12 | BIT14|BIT15;
	}
// debug
#if 1
	if(BeamformEntry.BeamformEntryState == BEAMFORMING_ENTRY_STATE_PROGRESSED)
	{
	
//	panic_printk("%d%s BW = %d\n",__LINE__,__FUNCTION__,BeamformEntry.BW);
		if(BeamformEntry.BW == HT_CHANNEL_WIDTH_20)
			BeamCtrlVal |= BIT9;
		else if(BeamformEntry.BW == HT_CHANNEL_WIDTH_20_40)
			BeamCtrlVal |= BIT10;
		else if(BeamformEntry.BW == HT_CHANNEL_WIDTH_80)
			BeamCtrlVal |= BIT11;
	}
#endif

	else
	{
		BeamCtrlVal &= ~(BIT9|BIT10|BIT11);
	}
	

//	PlatformEFIOWrite2Byte(Adapter, BeamCtrlReg, BeamCtrlVal);
	RTL_W16(BeamCtrlReg, BeamCtrlVal);

//	panic_printk("%s Idx %d BeamCtrlReg %x BeamCtrlVal %x, bw=%d\n", __FUNCTION__, Idx, BeamCtrlReg, BeamCtrlVal, BeamformEntry.BW);
}


//2REG_C2HEVT_CLEAR
#define		C2H_EVT_HOST_CLOSE			0x00	// Set by driver and notify FW that the driver has read the C2H command message
#define		C2H_EVT_FW_CLOSE			0xFF	// Set by FW indicating that FW had set the C2H command message and it's not yet read by driver.



VOID Beamforming_NDPARate_92E(
	struct rtl8192cd_priv *priv,
	BOOLEAN		Mode,
	u1Byte		BW,
	u1Byte		Rate)
{

#if 1
	u2Byte	NDPARate = Rate;

	if(NDPARate == 0)
	{
// ???	
#if 0
		if(Mode == 1 || Mode == 4)
			NDPARate = 0x0c;		//MGN_MCS0;
		else
			NDPARate = 0x2c;		//MGN_VHT1SS_MCS0;
#else
		if(priv->pshare->rssi_min > 30) // link RSSI > 30%
			NDPARate = 8;				//MGN_24M
		else
			NDPARate = 4;				//MGN_6M

#endif
	}

	if(NDPARate < 0x0c)
		BW = HT_CHANNEL_WIDTH_20;	

	RTL_W8(REG_NDPA_OPT_CTRL, (NDPARate<<2) |  (BW & 0x03));

#else
	{
		if(mode ==0) {
#if 1
			rate = 11;		// 54M
			BW = 0;
#else
			rate = 0x0c;		//_MCS0_RATE_;
			if(BW ==3)
				BW= 2;
#endif				
		}
		else {
			rate = 0x2c;		//_NSS1_MCS0_RATE_;
			
		}
	}
	RTL_W8(0x45f, (rate<<2) |  (BW & 0x03));
#endif	
}

// V1
#if 0
void TXBF_timer_callback(unsigned long task_priv)
{
	struct rtl8192cd_priv *priv = (struct rtl8192cd_priv *)task_priv;
	struct stat_info	*pstat;
	struct list_head	*phead, *plist;
	unsigned long flags;

	if(!priv->pmib->dot11RFEntry.txbf)
		return;

	SAVE_INT_AND_CLI(flags);

	phead = &priv->asoc_list;
	plist = phead->next;

//	panic_printk(".");

// VAP ?	
	
	while(plist != phead)
	{
		pstat = list_entry(plist, struct stat_info, asoc_list);

//		if(priv->pmib->dot11RFEntry.txbf == 1)
		if((pstat->ht_cap_len && (pstat->ht_cap_buf.txbf_cap)) ||
		(pstat->vht_cap_len && (cpu_to_le32(pstat->vht_cap_buf.vht_cap_info) & BIT(SU_BFEE_S))))

		{
//				BeamformingInit(priv, pstat->hwaddr, pstat->aid);
			if(pstat->vht_cap_len) //_eric_txbf
				BeamformingControl(priv, pstat->hwaddr, pstat->aid, 2, pstat->tx_bw);
			else
				BeamformingControl(priv, pstat->hwaddr, pstat->aid, 3, pstat->tx_bw);
		}


		if (plist == plist->next)
			break;
		plist = plist->next;

	};

	RESTORE_INT(flags);

	mod_timer(&priv->txbf_timer, jiffies + priv->pshare->rf_ft_var.soundingPeriod);

}
#endif

VOID
C2HTxBeamformingHandler_92E(
	struct rtl8192cd_priv *priv,
		pu1Byte			CmdBuf,
		u1Byte			CmdLen
)
{
	u1Byte 	status = (CmdBuf[0] & BIT0);
	
//	panic_printk("%d%s status = %d \n",__LINE__,__FUNCTION__,status);
	Beamforming_End(priv, status);
}

VOID
C2HTxTxReportHandler_92E(
	struct rtl8192cd_priv *priv,
		pu1Byte			CmdBuf,
		u1Byte			CmdLen
)
{
	struct tx_rpt rpt1;
	int k=0, j=0;
	for(j=0; j<2; j++) {
		rpt1.macid= CmdBuf[k];
		rpt1.txok = CmdBuf[k+1] | ((short)CmdBuf[k+2]<<8);
		rpt1.txfail = CmdBuf[k+3] | ((short)CmdBuf[k+4]<<8);
		rpt1.initil_tx_rate = CmdBuf[k+5];
		if(rpt1.macid != 0xff)
			txrpt_handler(priv, &rpt1);
		k+=6;
	}
}

VOID
_C2HContentParsing92E(
	struct rtl8192cd_priv *priv,
		u1Byte			c2hCmdId, 
		u1Byte			c2hCmdLen,
		pu1Byte 			tmpBuf
)
{
//	HAL_DATA_TYPE	*pHalData = GET_HAL_DATA(Adapter);
//	panic_printk("%d%s c2hCmdId = %d\n",__LINE__,__FUNCTION__,c2hCmdId);

	switch(c2hCmdId)
	{
/*
	case C2H_8812_DBG:
		RT_DISP(FC2H, C2H_Summary, ("[C2H], C2H_8812_DBG!!\n"));
		break;
	*/

	case C2H_8192E_TXBF:
	//	panic_printk("[C2H], C2H_8192E_TXBF!!\n");
		C2HTxBeamformingHandler_92E(priv, tmpBuf, c2hCmdLen);
		break;
	case C2H_8192E_TX_RATE:
		C2HTxTxReportHandler_92E(priv, tmpBuf, c2hCmdLen);
		break;

/*
	case C2H_8812_TX_REPORT:
		RT_DISP(FC2H, C2H_Summary, ("[C2H], C2H_8812_TX_REPORT!!\n"));
		C2HTxFeedbackHandler_8812(priv, tmpBuf, c2hCmdLen);
		break;

	case C2H_8812_RA_RPT:
		ODM_UpdateInitRate(&pHalData->DM_OutSrc, tmpBuf, c2hCmdLen);
		break;
*/
	
	default:
		break;
	}

}

VOID
C2HPacketHandler_92E(
	struct rtl8192cd_priv *priv,
		pu1Byte			Buffer,
		u1Byte			Length
	)
{
//	panic_printk("%d%s \n",__LINE__,__FUNCTION__);
	u1Byte	c2hCmdId=0, c2hCmdSeq=0, c2hCmdLen=0;
	pu1Byte tmpBuf=NULL;

	c2hCmdId = Buffer[0];
	c2hCmdSeq = Buffer[1];
	c2hCmdLen = Length -2;
	tmpBuf = Buffer+2;
	
	//panic_printk("[C2H packet], c2hCmdId=0x%x, c2hCmdSeq=0x%x, c2hCmdLen=%d\n", 
	//	c2hCmdId, c2hCmdSeq, c2hCmdLen);

//	RT_DISP_DATA(FC2H, C2H_ContentData, "[C2H packet], Content Hex:\n", tmpBuf, c2hCmdLen);
	_C2HContentParsing92E(priv, c2hCmdId, c2hCmdLen, tmpBuf);
//	panic_printk("%d%s Buffer = 0x%x\n",__LINE__,__FUNCTION__,tmpBuf);

	//PlatformEFIOWrite1Byte(Adapter, REG_C2HEVT_CLEAR_8812, C2H_EVT_HOST_CLOSE);
}

VOID HW_VAR_HW_REG_TIMER_START_92E(struct rtl8192cd_priv *priv)
{
	panic_printk("[%d][%s]\n",__LINE__,__FUNCTION__);
	RTL_W8(0x15F, 0x0);
	RTL_W8(0x15F, 0x5);

}

VOID HW_VAR_HW_REG_TIMER_INIT_92E(struct rtl8192cd_priv *priv, u2Byte t)
{
	panic_printk("[%d][%s]\n",__LINE__,__FUNCTION__);
	RTL_W8(0x164, 1);
	RTL_W16(0x15C, t);

}

VOID HW_VAR_HW_REG_TIMER_STOP_92E(struct rtl8192cd_priv *priv)
{
	panic_printk("[%d][%s]\n",__LINE__,__FUNCTION__);
	RTL_W8(0x15F, 0);
}
#endif


